;
;
; SS:SP = Return stack (push-down)
; SS:BP = Data stack (push-up)
;

%include "macros.inc"
%include "segments.inc"

SECTION .text



init
        mov     ax, STACKSEG    ;Setup segments
        mov     ss, ax
        mov     ax, SYSSEG
        mov     ds, ax
        mov     esp, STACKSIZE
        xor     ebp, ebp

startup
        mov     [_INBUF], cell DATA_END         ;Setup input buffer
        upsh    cell INBUF_SIZE
        call    ALLOT

        call    CR
        call    CR

        upsh    cell STARTUP_BLOCK              ;Load startup code
        call    LOAD



INTERPRET
        upsh    cell msg_ok
        upsh    cell msg_ok.len
        call    TYPE

        mov     [_TO_IN], cell 0
        mov     eax, [_INBUF]
        upsh    eax
        upsh    cell INBUF_SIZE
        call    ACCEPT

        upop    eax
        mov     [_INBUF_LEN], eax

        call    SPACE
        call    PARSE_INBUF
        jmp     INTERPRET
end INTERPRET



PARSE_INBUF
;
; Note: _INBUF_LEN is the *maximum* length of the input buffer. When reading
; from a block, it could be shorter. Therefore, stop at a NULL character.
;
        push    eax
.loop
        call    PARSE_WORD      ;new stack: c-addr u
        cmp     tosreg, cell 0
        jne     .find

        call    DROP
        call    DROP
        jmp     .end

.find
        call    FIND_WORD       ;new stack: c-addr u 0 | xt 1 | xt -1
        upop    eax
        cmp     eax, 0
        jz      .number         ;0 = failed
        jns     .immediate      ;1 = Immediate word: don't compile

        ;Not immediate. Compile it?
        cmp     [_STATE], cell 0
        jnz     .compile

.immediate      ;Just run it!
        upop    eax
        call    eax
        jmp     .loop

.compile
        call    COMPILE
        jmp     .loop

.number
        call    TO_NUMBER
        jc      .error
        cmp     [_STATE], cell 0        ;Compile it?
        je      .loop   ;No
        call    LITERAL ;Yes - have LITERAL take care of it
        jmp     .loop

.error
        call    CR
        upsh    cell '"'
        call    EMIT
        call    TYPE
        upsh    cell '"'
        call    EMIT
        upsh    cell msg_undef
        upsh    cell msg_undef.len
        call    TYPE

        ;Error recovery: Reset state & clear stack
        mov     [_STATE], cell 0
        mov     [_SOURCE_ID], cell 0
        xor     ebp, ebp
.end
        pop     eax
        ret
end PARSE_INBUF



PARSE_WORD      ;( "<spaces>name" -- c-addr u , parse a space-delimited word )
        push    eax
        push    ebx
        push    esi

        mov     ebx, [_INBUF]
        add     ebx, [_TO_IN]

        mov     eax, [_INBUF_LEN]
        sub     eax, [_TO_IN]

        ;Skip spaces
        mov     esi, -1
.skipspace
        inc     esi
        cmp     esi, eax
        jge     .s1
        cmp     byte [ebx+esi], 0x20    ;SPACE
        je      .skipspace
        cmp     byte [ebx+esi], 0x09    ;TAB
        je      .skipspace
        cmp     byte [ebx+esi], 0x0D    ;CR
        je      .skipspace
        cmp     byte [ebx+esi], 0x0A    ;LF
        je      .skipspace
.s1
        add     [_TO_IN], esi
        sub     eax, esi

        add     ebx, esi
        upsh    ebx             ;Push address

        ;Find length
        sub     esi, esi
.length
        cmp     esi, eax                ;End of command line reached?
        jge     .end
        cmp     byte [ebx+esi], 0       ;NULL
        je      .end
        cmp     byte [ebx+esi], 0x20    ;SPACE
        je      .end
        cmp     byte [ebx+esi], 0x09    ;TAB
        je      .end
        cmp     byte [ebx+esi], 0x0D    ;CR
        je      .end
        cmp     byte [ebx+esi], 0x0A    ;LF
        je      .end
        inc     esi
        jmp     .length

.end
        upsh    esi             ;Push length
        inc     esi
        add     [_TO_IN], esi

        pop     esi
        pop     ebx
        pop     eax
        ret
end PARSE_WORD



FIND_WORD       ;( c-addr u -- c-addr u 0 | xt 1 | xt -1 )
;
; Looks up a string in the dictionary.
; Return flags:
;        0 = not found (string is left on stack)
;        1 = immediate
;       -1 = normal
;
        push    eax
        push    ebx
        push    ecx
        push    edx

        upop    edx     ;Length
        upop    eax     ;Address

        call    GET_ORDER
        upop    ecx

.loop
        upsh    eax
        upsh    edx
        call    ROT
        call    SEARCH_WORDLIST
        upop    ebx
        cmp     ebx, 0
        jne     .end
        loop    .loop

        ;Failed
        upsh    eax
        upsh    edx
.end
        upsh    ebx

        pop     edx
        pop     ecx
        pop     ebx
        pop     eax
        ret
end FIND_WORD



SEARCH_WORDLIST ;( c-addr u wid -- 0 | xt 1 | xt -1 )
        push    eax
        push    ebx
        push    ecx
        push    edx
        push    esi
        push    edi

        upop    ebx
        mov     ebx, [ebx]
        upop    edx                     ;Length (must be <128)
.loop
        cmp     ebx, 0                  ;Done?
        je      .failed
        mov     al, [ebx+CELLSIZE]
        and     al, 0x1F                ;Strip the 3 flag bits
        cmp     al, dl
        je      .same_length
.next
        mov     ebx, [ebx]              ;Get next entry
        jmp     .loop

.failed
        call    DROP
        upsh    cell 0
        jmp     .end

.same_length
        tos     esi                     ;Address of query string
        mov     edi, CELLSIZE+1         ;Address of string in dictionary
        mov     ecx, edx
.compare
        mov     al, [ebx+edi]           ;AL = dict
        mov     ah, [esi]               ;AH = query

        ;Convert AL to uppercase
        cmp     al, 'a'
        jb      .c1
        cmp     al, 'z'
        ja      .c1
        sub     al, 32
.c1
        ;Convert AH to uppercase
        cmp     ah, 'a'
        jb      .c2
        cmp     ah, 'z'
        ja      .c2
        sub     ah, 32
.c2

        cmp     al, ah
        jne     .next
        inc     esi
        inc     edi
        loop    .compare

        ;Success
        call    DROP
        add     edi, ebx
        mov     edi, [edi]              ;Get CFA
        upsh    edi
        test    [ebx+CELLSIZE], byte 0x80
        jnz     .immediate

        ;Normal (not immediate)
        upsh    cell -1
        jmp     .end
.immediate
        upsh    cell 1

.end
        pop     edi
        pop     esi
        pop     edx
        pop     ecx
        pop     ebx
        pop     eax
        ret
end SEARCH_WORDLIST



TO_NUMBER       ;( c-addr u -- n|u )
;
; On error, returns with CF set (and nothing on stack)
;
        push    eax
        push    ebx
        push    ecx
        push    edx
        push    esi
        push    edi

        upop    ecx             ;Count
        upop    ebx             ;String
        xor     esi, esi        ;Index into string
        xor     eax, eax        ;AX = the number
        xor     edx, edx        ;Temp
        xor     edi, edi        ;DI = -1 if number is negative, 0 if positive

        cmp     [ebx], byte '-'
        jne     .loop
        dec     edi             ;Set negative flag (handle it later)
        inc     esi

.loop
        mov     dl, [ebx+esi]
        cmp     dl, '0'         ;Sort out the digit
        jb      .bad
        cmp     dl, '9'
        jbe     .decimal
        cmp     dl, 'A'
        jb      .bad
        cmp     dl, 'Z'
        jbe     .upr
        cmp     dl, 'a'
        jb      .bad
        cmp     dl, 'z'
        jbe     .lwr
        jmp     .bad
.lwr
        sub     dl, 'a'-10      ;Convert a lowercase letter to a number
        jmp     .check
.upr
        sub     dl, 'A'-10      ;Convert an uppercase letter to a number
        jmp     .check
.decimal
        sub     dl, '0'         ;Convert a decimal digit to a number

.check
        cmp     dl, byte [_BASE]
        jae     .bad
.add
        push    edx
        mul     word [_BASE]
        pop     edx
        add     eax, edx

        inc     esi             ;Next char?
        cmp     esi, ecx
        jb      .loop
.done
        cmp     edi, 0
        jz      .positive
        neg     eax
.positive
        upsh    eax
        clc
        jmp     .end
.bad
        upsh    ebx
        upsh    ecx
        stc
.end
        pop     edi
        pop     esi
        pop     edx
        pop     ecx
        pop     ebx
        pop     eax
        ret
end TO_NUMBER



LITERAL ;( x -- , Compile an instruction to push x on the stack )
        ;Compile-only word!

        ;Generate MOV [BP], imm
        upsh    cell 0xC7       ;C7 = mov r/m, imm
        call    C_COMMA
        upsh    cell 0x46       ;mod r/m = 01 000 110 (SS:BP+d8)
        call    C_COMMA
        upsh    cell 0x00       ;d8 = 0 (no displacement)
        call    C_COMMA
        call    COMMA           ;Add x to the instruction

        ;Generate ADD BP, CELLSIZE
        upsh    cell 0x81
        call    C_COMMA
        upsh    cell 0xC5
        call    C_COMMA
        upsh    cell CELLSIZE
        call    COMMA
        ret
end LITERAL



COMMA
        push    eax
        push    edi

        upop    eax
        mov     edi, [_DSP]
        mov     [edi], eax
        add     cell [_DSP], CELLSIZE

        pop     edi
        pop     eax
        ret
end COMMA



C_COMMA
        push    eax
        push    edi

        upop    eax
        mov     edi, [_DSP]
        mov     [edi], al
        inc     cell [_DSP]

        pop     edi
        pop     eax
        ret
end C_COMMA



; colon-sys is: <start address>

NEW_WORD        ;( "<spaces>name" -- colon-sys , Create new word entry )
        push    eax
        push    ebx

        call    PARSE_WORD
        tos     eax
        cmp     eax, 0
        jz      .error

        push    cell [_DSP]             ;Save start address

        mov     ebx, [comp_wordlist]
        mov     ebx, [ebx]
        upsh    ebx
        call    COMMA                   ;Add link to next word

        call    DUP
        call    C_COMMA                 ;Add length

        call    SWAP
        call    OVER
        mov     eax, [_DSP]
        upsh    eax
        call    SWAP
        call    _CMOVE                  ;Add name
        call    ALLOT

        mov     eax, [_DSP]
        add     eax, 2*CELLSIZE
        upsh    eax
        call    COMMA                   ;Add CFA
        upsh    cell 0
        call    COMMA                   ;Add code size (0 for now)

        ;Push colon-sys
        pop     eax
        upsh    eax                     ;Start address
.end
        pop     ebx
        pop     eax
        ret
.error
        call    DROP
        call    DROP
        upsh    cell .msg_err
        upsh    cell .msg_err_len
        call    TYPE
        jmp     .end

.msg_err        db "name expected",13,10
.msg_err_len    equ     $ - .msg_err
end NEW_WORD



COMPILE ;( xt -- , Compile a "CALL xt" instruction )
        upsh    cell 0xE8       ;Generate a CALL instruction
        call    C_COMMA

        ;Convert xt to a relative address
        ; displacement = dest origin - = dest HERE CELL+ -
        call    HERE
        call    CELL_PLUS
        call    MINUS

        call    COMMA           ;Destination address
        ret
end COMPILE



START_COMPILE
        call    NEW_WORD
        mov     cell [_STATE], -1
        ret
end START_COMPILE



END_COMPILE     ;( colon-sys -- )
        mov     cell [_STATE], 0
        call    ADD_WORD
        ret
end END_COMPILE



ADD_WORD        ;Add new word to dictionary (after defining it)
        push    eax
        push    ebx

        upsh    cell 0xC3               ;Add RET instruction
        call    C_COMMA

        ;Need to set code size

        ;Link into dictionary
        mov     ebx, [comp_wordlist]    ;Add new entry to wordlist
        upop    eax
        mov     [ebx], eax

        pop     ebx
        pop     eax
        ret
end ADD_WORD



IMMEDIATE       ;( -- , Make the last word defined IMMEDIATE )
        push    ebx

        mov     ebx, [comp_wordlist]
        mov     ebx, [ebx]
        add     ebx, CELLSIZE
        or      byte [ebx], 0x80

        pop     ebx
        ret
end IMMEDIATE



POSTPONE        ;( "<spaces>name" -- )
                ;Compile the next word even if it's IMMEDIATE
        push    eax
.loop
        call    PARSE_WORD      ;new stack: c-addr u
        cmp     tosreg, cell 0
        jne     .find

        call    DROP
        call    DROP
        jmp     .end

.find
        call    FIND_WORD       ;new stack: c-addr u 0 | xt 1 | xt -1
        upop    eax
        cmp     eax, 0
        jz      near    PARSE_INBUF.number         ;0 = failed
        jmp     PARSE_INBUF.compile
.end
        ret
end POSTPONE



CONSTANT        ; ( x "<spaces>name" -- )
        call    NEW_WORD
        call    SWAP
        call    LITERAL
        call    ADD_WORD
        ret
end CONSTANT



VARIABLE        ; ( "<spaces>name" -- )
        call    HERE
        upsh    cell CELLSIZE
        call    ALLOT

        call    NEW_WORD
        call    SWAP
        call    LITERAL
        call    ADD_WORD
        ret
end VARIABLE



CREATE
        call    NEW_WORD
        call    HERE
        upsh    cell 6 + 2*CELLSIZE
        call    PLUS
        call    LITERAL
        call    ADD_WORD
        ret
end CREATE



DOES
        ret
end DOES



%include "block.asm"
%include "flow.asm"
%include "graphics.asm"
%include "keyboard.asm"
%include "logic.asm"
%include "math.asm"
%include "misc.asm"
%include "stack.asm"
%include "text.asm"

%include "dict.inc"



GLOBAL_VARS

_DSP            dc      DATA_END        ;'HERE' (data space pointer)
_BASE           dc      10
_SOURCE_ID      dc      0
_BLK            dc      0
_STATE          dc      0

;Messages
msg_ok          db      " Ok",13,10
        .len    equ     $ - msg_ok
msg_undef       db      " undefined",13,10
        .len    equ     $ - msg_undef

;Unitialized data
SECTION .bss

_INBUF          resc    1
_INBUF_LEN      resc    1
_TO_IN          resc    1

end GLOBAL_VARS

DATA_END

